#!/usr/bin/env python
#syntax that makes sense
from __future__ import division, print_function
#manipulate filesystem paths, set env variable
import os
#communicate with Shell.FM
import dbus, subprocess
#gui and application loop stuff
import gtk, glib, gobject, pango
from pygtkhelpers import gthreads
from dbus.mainloop.glib import DBusGMainLoop
#open urls with default browser
import webbrowser
#save application data between runtimes
import shelve
#multimedia key support
import keybinder
#ensure single-instance
import fcntl, sys
#lastfm functions
import pylast
#download cover art, other info
import urllib
#notifications
import pynotify
#filter html for the text content
import re


def filter_html_tags(html_str):
    output = html_str
    regex = re.compile('<[^>]*>')
    for substring in regex.findall(html_str):
        output = output.replace(substring, '')
    return output

def extract(source, intro, outro):
    start = source.index(intro) + len(intro)
    end = source.index(outro, start)
    return source[start:end]

icon_map = {'library': 'emblem-sound',
'recommended': 'face-wink',
'neighborhood': 'system-users',
'loved': 'emblem-favorite',
'artist': 'audio-input-microphone',
'tag': 'tomboy',
'group': 'system-users',
'on-the-fly': 'system-users',
'friend': 'user-available',
'neighbor': 'user-available'}

class ToolButtonStation(gtk.ToolButton):

    def __init__(self, station, label=None):
        self.station = station
        if not label:
            label = self.station.name
        super(ToolButtonStation, self).__init__(None, label)
        self.set_icon_name(icon_map[self.station.category])
        self.set_tooltip_text('Play Station')

class MenuItemStation(gtk.ImageMenuItem):

    def __init__(self, station, label=None):
        self.station = station
        if not label:
            label = self.station.name
        super(MenuItemStation, self).__init__()
        self.set_label(label)
        icon_name = icon_map[self.station.category]
        image = gtk.image_new_from_icon_name(icon_name, gtk.ICON_SIZE_MENU)
        self.set_image(image)
        self.set_tooltip_text('Play Station')

class Service():
    # Common interface for Lastfm, SomaFM, Pandora, etc.
    def get_total_seconds(self):
        return 0

    def get_remaining_seconds(self):
        return 0

    def play_pause_track(self):
        return False

    def play_station(self, station):
        return False

    def stop_track(self):
        pass

    def ban_track(self):
        pass

    def login(self, username='', password=''):
        pass

    def connect(self):
        pass

    def disconnect(self):
        pass

    def love_track(self):
        pass

    def skip_track(self):
        pass

    def get_track(self):
        return ''

    def get_artist(self):
        return ''

    def get_album(self):
        return ''

    def get_artist_url(self):
        return ''

    def get_album_url(self):
        return ''

    def get_track_url(self):
        return ''

    def get_station(self, type, subject, category):
        return None

class Lastfm(Service):

    def __init__(self, username='<username>', password='<password>',
    host='localhost', port='54311'):
        self.service_id = 'lastfm'
        self.api_key = '04d01b6a9ed86c4eb893dc585ea8d7c4'
        self.api_secret = '9c98b509274bd62a0264f27e50bed381'
        self.username = username
        self.password = password
        self.host = host
        self.port = port
        home_path = os.path.expanduser('~')
        self.chipmunk_path = os.path.join(home_path, '.chipmunk')
        os.putenv('SHELL_FM_HOME', self.chipmunk_path)
        if not os.path.exists(self.chipmunk_path):
            os.makedirs(self.chipmunk_path)
        self.shellfm_config_path = os.path.join(self.chipmunk_path, 'shell-fm.rc')
        self.connected = False
        if username != '<username>':
            self.login(self.username, self.password)

    def clobber_shellfm_config(self):
        lines = ['username = '+self.username,
        'password = '+self.password,
        'np-cmd = dbus-send /org/bitbucket/chipmunk org.bitbucket.chipmunk.shellfm_track_started']
        lines = '\n'.join(lines)
        f = open(self.shellfm_config_path, 'w')
        f.write(lines)
        f.close()

    def login(self, new_username='<username>', new_password='<password>'):
        self.username = new_username
        self.password = new_password
        self.clobber_shellfm_config()
        return self.connect()

    def disconnect(self):
        subprocess.call(['pkill', 'shell-fm'])
        subprocess.call(['pkill', '-9', 'shell-fm'])

    def shellfm_is_running(self):
        proc = subprocess.Popen(['pgrep', 'shell-fm'], stdout=subprocess.PIPE)
        output = proc.communicate()[0]
        return output != ''

    def connect(self):
        self.disconnect()
        subprocess.call(['shell-fm', '-d', '-i', self.host, '-p', self.port])
        try:
            self.network = self.get_pylast_network()
            self.connected = True
        except:
            self.network = None
            self.connected = False
        return self.connected

    def get_pylast_network(self):
        return pylast.get_lastfm_network(api_key=self.api_key,
        api_secret=self.api_secret,
        username=self.username,
        password_hash=pylast.md5(self.password))

    def share_track(self, user_list, message=None):
        track = self.network.get_track(self.get_artist(), self.get_track())
        if message == '':
            message = None
        track.share(user_list, message)

    def shout_track(self, user_list, message=''):
        artist, title = self.get_artist(), self.get_track()
        track = '[track artist={a}]{t}[/track]'.format(a=artist, t=title)
        message += '  \n' + track
        for user in user_list:
            if '@' not in user:#only shout to last.fm users, not email addresses.
                self.network.get_user(user).shout(message)

    def is_subscriber(self):
        try:
            return self.network.get_user(self.username).is_subscriber()
        except:
            return False

    def shellfm(self, command):
        proc = subprocess.Popen(['nc.openbsd', '-w 3', self.host, self.port],
        stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        #Using socat may cause timing issues that I don't understand
        #addr = ':'.join(['tcp', self.host, self.port])
        #proc = subprocess.Popen(['socat', '-t3', '-', addr],
        #stdin=subprocess.PIPE, stdout=subprocess.PIPE, stderr=subprocess.PIPE)

        try:
            output, error = proc.communicate(command)
            output = output.strip()
        except:
            output, error = '', ''
            print('Aha! This except clause IS useful!')
        else:
            try:
                if output[0] == '"' and output[-1] == '"':
                    output = output[1:-1]#removes surrounding quotes
            except:
                pass
            if error:
                print('shell-fm complaint:', error)
#                if 'echo: write error: broken pipe' in error.lower():
#                    print('Connection to shell-fm backend lost. Restarting it...')
#                    self.connect()
#                    print('Now resuming last attempted operation ('+command+')...')
#                    return self.shellfm(command)
        return output

    def play_pause_track(self):
        #if self.shellfm_is_running():
        self.shellfm('pause')
        return self.get_track().strip() != ''

    def skip_track(self):
        self.shellfm('skip')

    def love_track(self):
        self.shellfm('love')

    def ban_track(self):
        self.shellfm('ban')

    def stop_track(self):
        self.shellfm('stop')

    #def tag_artist(self, tag_list):#works?
        #self.shellfm('tag-artist ' + ','.join(tag_list))

    #def tag_album(self, tag_list):#works?
        #self.shellfm('tag-album ' + ','.join(tag_list))

    def tag_track(self, tag_list):#works?
        self.shellfm('tag-track ' + ','.join(tag_list))

    #def get_artist_tags(self):#worky?
        #return self.shellfm('artist_tags')

    #def get_album_tags(self):#worky?
        #return self.shellfm('album_tags')

    def get_top_tags(self, artist, title, limit):
        try:
            track = self.network.get_track(artist, title)
            top_tag_items = track.get_top_tags(limit)
            return sorted([str(top_item.item) for top_item in top_tag_items], key=str.lower)
        except:
            return []

    def get_friends(self, limit):
        try:
            friends = self.network.get_user(self.username).get_friends(limit)
            return sorted([str(friend) for friend in friends], key=str.lower)
        except:
            return []

    #def get_groups(self, limit):
        #try:
            #groups = self.network.get_user(self.username).get_groups(limit)
            #return sorted([str(group) for group in groups], key=str.lower)
        #except:
            #return []

    def get_neighbors(self, limit):
        try:
            neighbors = self.network.get_user(self.username).get_neighbours(limit)
            return sorted([str(neighbor) for neighbor in neighbors], key=str.lower)
        except:
            return []

    def get_artist_bio(self, artist, summary):
        try:
            if summary:
                bio = self.network.get_artist(artist).get_bio_summary()
            else:
                bio = self.network.get_artist(artist).get_bio_content()
            return filter_html_tags(bio)
        except:
            return ''

    def get_venue_name(self, event):

        def extract(source, intro, outro):
            #how exactly do the results of this method differ from those of the global one? Is this one completely redundant?
            regex = re.compile(intro + '.*' + outro)
            match = regex.search(source).group()
            return match[len(intro) : 0 - len(outro)]

        url = 'http://ws.audioscrobbler.com/2.0/?method=event.getinfo&event={e}&api_key={k}'
        url = url.format(e=str(event.get_id()), k=self.api_key)
        event_xml = urllib.urlopen(url).read()
        name = extract(event_xml, '<name>', '</name>')
        city = extract(event_xml, '<city>', '</city>')
        country = extract(event_xml, '<country>', '</country>')
        return ', '.join([name, city, country])

    def get_artist_events(self, artist):
        try:
            events = self.network.get_artist(artist).get_upcoming_events()
            event_dictionaries = []
            for event in events:
                event_dict = {}

                try:
                    event_dict['begins'] = str(event.get_start_date())
                except:
                    event_dict['begins'] = ''
                if event_dict['begins'].strip() == '':
                    continue

                try:
                    event_dict['title'] = str(event.get_title())
                except:
                    event_dict['title'] = self.chipmunk.current_artist

                try:
                    event_dict['venue'] = self.get_venue_name(event)
                    #event_dict['venue'] = str(event.get_venue().get_name())
                except:
                    event_dict['venue'] = ''

                try:
                    event_dict['description'] = filter_html_tags(str(event.get_description()))
                except:
                    event_dict['description'] = ''
                if event_dict['description'] in [None, 'None']:
                    event_dict['description'] = ''

                try:
                    event_dict['headliner'] = str(event.get_headliner())
                except:
                    event_dict['headliner'] = ''

                event_dict['url'] = event.get_url()

                event_dictionaries.append(event_dict)
            return event_dictionaries
        except:
            return []

    def get_track_wiki(self, artist, title, summary=False):
        try:
            if summary:
                wiki = self.network.get_track(artist, title).get_wiki_summary()
            else:
                wiki = self.network.get_track(artist, title).get_wiki_content()
            return filter_html_tags(wiki)
        except:
            return ''

    def get_cover_image_url(self):
        #Maybe this is the same as using pylast, like I did in the beginning.
        #I switched from that to the shellfm call, since I wouldn't have to
        #loop through the sizes and it was simple and fast.
        #Now I realize that shellfm gives a tiny cover when I can get a huge one.
        #The lastfm url api is really easy to use, not sure if I should replace this with a pylast call.
        url = 'http://ws.audioscrobbler.com/2.0/?method=album.getinfo&api_key={k}&artist={ar}&album={al}'
        url = url.format(k=self.api_key, ar=self.get_artist(), al=self.get_album())
        album_xml = urllib.urlopen(url).read()
        for size in ['extralarge', 'large', 'medium', 'small']:
            try:
                return extract(album_xml, '<image size="'+size+'">', '</image>')
            except:
                pass
        return self.shellfm('info "%I"')

    def get_artist(self):
        return self.shellfm('info "%a"')

    def get_volume(self):
        return self.shellfm('info "%v"')

    def set_volume(self, new_perc):
        new_vol = (new_perc/100) * 64
        return self.shellfm('volume '+str(new_vol))

    def get_track(self):
        return self.shellfm('info "%t"')

    def get_album(self):
        return self.shellfm('info "%l"')

    def get_total_seconds(self):
        try:
            return int(self.shellfm('info "%d"'))
        except:
            return 0

    def get_station_name(self):
        return self.shellfm('info "%s"')

    def get_station_url(self):
        return self.shellfm('info "%S"')

    def get_station(self, station_type=None, subject=None, category=None):

        def urlify(station_type, subject):
            base = 'lastfm://'
            user_type_map = {'library': '/personal',
            'recommended': '/recommended',
            'neighborhood': '/neighbours',
            'loved': '/loved'}
            general_type_map = {'artist': 'artist/',
            'tag': 'globaltags/',
            'group': 'group/'}
            if station_type in user_type_map:
                return base + 'user/' + subject + user_type_map[station_type]
            else:
                return base + general_type_map[station_type] + subject

        if not station_type:
            name = self.get_station_name()
            url = self.get_station_url()
        else:
            #subject = subject.title()#this would result in "Andy'S 80S Mix", so instead:
            if type(subject) == type(''):
                subject = subject.strip().lower()
                formatted = ' '
                for char in subject:
                    if formatted[-1] in [' ', '/', '.']:#letters after these characters are capitalized
                        formatted += char.upper()
                    else:
                        formatted += char
                subject = formatted.strip()
            if type(subject) == type([]):
                category = 'on-the-fly'
                name = ', '.join(subject)
                url = [urlify(station_type, x) for x in subject]
            else:
                if not category:
                    category = station_type
                name = subject
                url = urlify(station_type, subject)
        return Station(self.service_id, name, url, category)

    def get_album_url(self):
        return self.shellfm('info "%L"')

    def get_artist_url(self):
        return self.shellfm('info "%A"')

    def get_track_url(self):
        return self.shellfm('info "%T"')

    def get_remaining_seconds(self):
        try:
            return int(self.shellfm('info "%R"'))
        except:
            return 0

    def toggle_discovery_mode(self):
        new_state = self.shellfm('discovery')
        if new_state == 'DISCOVERY ON':
            return True
        elif new_state == 'DISCOVERY OFF':
            return False
        return None

    def toggle_scrobble(self):
        new_state = self.shellfm('rtp')
        if new_state == 'RTP ON':
            return True
        elif new_state == 'RTP OFF':
            return False
        return None

    #def download_track(self, destination):
        #pass

    def play_station(self, station):
        if not self.shellfm_is_running():
            self.login()
        else:
            self.shellfm('play '+station.url)
        #return self.get_track().strip() != ''
        max_attempts = 10
        if self.get_track().strip() == '':
            for attempt in range(max_attempts):
                if self.get_track().strip() != '':
                    return True
            return False
        return True

class Station():

    def __init__(self, service_id, name, url, category=None):
        self.service_id = service_id
        self.name = name
        self.category = category
        if type(url) == type([]):
            self.url_list = url
            self.url = self.url_list[0]
        else:
            self.url_list = None
            self.url = url
        self.image = None

    def rotate(self):
        if self.url_list:
            self.url_list = self.url_list[1:] + self.url_list[:1]
            self.url = self.url_list[0]
            return True
        return False

    def __eq__(self, other):
        try:
            if other == None:
                return False
        except:
            pass
        if self.category == 'on-the-fly' and other.category == 'on-the-fly':
            if set(self.url_list) == set(other.url_list):
                return True
        properties = [(self.service_id, other.service_id),
        (self.name.lower(), other.name.lower()),
        (self.category, other.category)]
        for a, b in properties:
            if a != b:
                return False
        return True

    def __ne__(self, other):
        return not self.__eq__(other)

    def __str__(self):
        return self.name

class Chipmunk():
    #this class manages most of the radio functionality,
    #while ChipmunkApp manages most of the gtk stuff
    #but the distinction is not that clearly defined
    #e.g. Chipmunk handles notifications, ChipmunkApp does some queue handling
    def __init__(self):
        self.config_dir = os.path.join(os.path.expanduser('~'), '.chipmunk')
        if not os.path.isdir(self.config_dir):
            os.mkdir(self.config_dir)
        self.db = os.path.join(self.config_dir, 'chipmunk_data')
        self.services = {}
        self.services['lastfm'] = Lastfm()
        self.current_station = None
        self.current_artist = ''
        self.current_track = ''
        self.cover_image_url = ''
        self.queued_task = None
        self.status = 'stopped'
        self.time = 0

    def play_file(self, file_path):
        self.services['lastfm'].play_file(file_path)

    def get_station(self, station_type=None, subject=None, category=None, service_id='lastfm'):
        return self.services[service_id].get_station(station_type, subject, category)

    def is_subscriber(self):
        service = self.services['lastfm']
        return service.is_subscriber()

    def toggle_discovery_mode(self, service_id='lastfm'):
        if service_id == 'lastfm':
            mode = self.services[service_id].toggle_discovery_mode()
            #self.save('discovery', mode)
            return mode

    def toggle_scrobble(self, service_id='lastfm'):
        if service_id == 'lastfm':
            return self.services[service_id].toggle_scrobble()

    def save(self, key, value):
        shelf = shelve.open(self.db)
        shelf[key] = value
        shelf.close()

    def get_username(self, service_id='lastfm'):
        return self.services[service_id].username

    def get_password(self, service_id='lastfm'):
        return self.services[service_id].password

    def load(self, key):
        shelf = shelve.open(self.db)
        try:
            value = shelf[key]
        except:
            default_map = {'lastfm_username': '<username>',
            'lastfm_password': '<password>', 'auto_fit': False,
            'hide_menubar': False, 'cover_size': 96, 'label_album_scale': 1,
            'recent_stations': [], 'recent_artists': [], 'recent_tags': [],
            'recent_groups': [], 'window_size': (775, 1), 'bookmarks': [],
            'shout_when_share': True, 'is_subscriber': False,
            'show_context': False, 'label_title_scale': 2.5, 'discovery': False,
            'label_artist_scale': 1.5, 'small_icons': True}
            try:
                value = default_map[key]
            except:
                value = None
        shelf.close()
        return value

    def play_pause_track(self):
        service = self.services[self.current_station.service_id]
        if not service.play_pause_track():
            return self.play_station(self.current_station)
        elif self.status == 'playing':
            self.status = 'paused'
        else:
            self.status = 'playing'
        return True

        #alternative thought
        #service = self.services[self.current_station.service_id]
        #if self.status == 'stopped':
            #return self.play_station(self.current_station)
        #else:
            #if self.status == 'playing':
                #self.status = 'paused'
            #else:
                #self.status = 'playing'
            #return service.play_pause_track()

    def notify(self, message=None, sub=None, icon='chipmunk-radio', timeout=5000):
        if message == None:
            title = self.current_track
            body = self.current_artist
        else:
            title = message
            body = sub
        try:
            notification = pynotify.Notification(title, body, icon)
            notification.set_timeout(timeout)
            #notification.set_urgency(pynotify.URGENCY_NORMAL)
            notification.show()
        except:
            pass

    def get_remaining_seconds(self):
        return self.services[self.current_station.service_id].get_remaining_seconds()

    def get_total_seconds(self):
        return self.services[self.current_station.service_id].get_total_seconds()

    def play_station(self, station):
        self.current_station = station
        return self.services[station.service_id].play_station(station)

    def stop_track(self):
        self.services[self.current_station.service_id].stop_track()
        self.status = 'stopped'

    def ban_track(self):
        self.notify('Track Banned', self.current_track, 'emblem-unreadable')
        self.services[self.current_station.service_id].ban_track()

    def share_track(self, user_list, message=None):
        service = self.services[self.current_station.service_id]
        service.share_track(user_list, message)

    def shout_track(self, user_list, message=''):
        service = self.services[self.current_station.service_id]
        service.shout_track(user_list, message)

    def login(self, service_id='lastfm', new_username=None, new_password=None):
        if not new_username:
            new_username = self.load(service_id + '_username')
        if not new_password:
            new_password = self.load(service_id + '_password')
        self.save(service_id + '_username', new_username)
        self.save(service_id + '_password', new_password)
        service = self.services[service_id]
        service.login(new_username, new_password)
        if service_id == 'lastfm':
            self.current_station = self.get_station('library', service.username)

    def disconnect(self, service_id='lastfm'):
        self.services[service_id].disconnect()

    def get_cover_image(self):
        url = self.cover_image_url
        if not url:
            return None
        filepath = os.path.join(self.config_dir, 'cover.jpg')
        image = urllib.URLopener()
        try:
            image.retrieve(url, filepath)
            return filepath
        except:
            return None

    def get_cover_image_url(self):
        service = self.services[self.current_station.service_id]
        return service.get_cover_image_url()

    def clear_queue(self):
        self.queued_task = None

    def process_queue(self):
        if not self.queued_task:
            return False
        task = self.queued_task
        self.clear_queue()
        task[0](*task[1])
        if task[0] == self.stop_track:
            self.notify('Stopped', '', 'media-playback-stop')
        return True

    def queue_station(self, station):
        self.queued_task = (self.play_station, [station])

    def queue_stop(self):
        self.queued_task = (self.stop_track, [])

    def love_track(self):
        service = self.services[self.current_station.service_id]
        service.love_track()
        self.notify('Track Loved', self.current_track, 'emblem-favorite')

    def skip_track(self):
        service = self.services[self.current_station.service_id]
        service.skip_track()
        if self.status == 'paused':
            self.play_pause_track()

    def get_track(self):
        service = self.services[self.current_station.service_id]
        return service.get_track()

    def get_top_tags(self, limit=3):
        try:
            service = self.services[self.current_station.service_id]
            return service.get_top_tags(self.current_artist, self.current_track, limit)
        except:
            return []

    def get_top_tag(self):
        try:
            return self.get_top_tags(1)[0]
        except:
            return None

    def get_friends(self, limit=100):
        try:
            service = self.services[self.current_station.service_id]
            return service.get_friends(limit)
        except:
            return []

    def get_neighbors(self, limit=50):
        try:
            service = self.services[self.current_station.service_id]
            return service.get_neighbors(limit)
        except:
            return []

    def get_artist(self):
        service = self.services[self.current_station.service_id]
        return service.get_artist()

    def get_album(self):
        service = self.services[self.current_station.service_id]
        return service.get_album()

    def get_album_url(self):
        service = self.services[self.current_station.service_id]
        return service.get_album_url()

    def get_artist_url(self):
        service = self.services[self.current_station.service_id]
        return service.get_artist_url()

    def get_volume(self):
        service = self.services[self.current_station.service_id]
        return service.get_volume()

    def set_volume(self, vol):
        service = self.services[self.current_station.service_id]
        service.set_volume(vol)

    def get_track_url(self):
        service = self.services[self.current_station.service_id]
        return service.get_track_url()

    def get_lyrics(self):

        def get_lyrics_helper(artist, title, engine):
            try:
                if artist and title:
                    return engine(artist, title)
                else:
                    return ''
            except:
                return ''

        engines = [self.get_lyrics_lyricstime, self.get_lyrics_lyricsvip]
        for engine in engines:
            lyrics = get_lyrics_helper(self.current_artist, self.current_track, engine)
            if lyrics != '':
                return lyrics
        return ''

    #def get_lyrics_chartlyrics(self, artist, title):
        #url = 'http://api.chartlyrics.com/apiv1.asmx/SearchLyricDirect?artist='+artist[:75]+'&song='+title[:125]
        #lyrics = urllib.urlopen(url).read()
        #lyrics = extract(lyrics, '<Lyric>', '</Lyric>').strip()
        #if lyrics != '':
            #lyrics += '\n\nSource: http://www.chartlyrics.com'
        #return lyrics

    def get_lyrics_lyricsvip(self, artist, title):
        url = 'http://www.lyricsvip.com/'+artist+'/'+title+'-Lyrics.html'
        url = url.replace(' ', '-')
        lyrics = extract(urllib.urlopen(url).read(), '</h2>', '&nbsp;')
        lyrics = filter_html_tags(lyrics).strip()
        special_chars = {'&quot;': '"', '&amp;': '&'}
        for bad, good in special_chars.items():
            lyrics = lyrics.replace(bad, good)
        if lyrics != '':
            lyrics += '\n\nSource: http://www.lyricsvip.com'
        return lyrics

    def get_lyrics_lyricstime(self, artist, title):
        if artist[-1] == "'":
            artist = artist[:-1]
        if title[-1] == "'":
            title = title[:-1]
        url = artist+'-'+title+'-lyrics'
        for char in ['!', ',', '(', ')', '&', '?']:
            url = url.replace(char, '')
        for char in ['. ', '.', '  ', ' ', "'"]:
            url = url.replace(char, '-')
        url = 'http://www.lyricstime.com/' + url.lower() + '.html'
        lyrics = extract(urllib.urlopen(url).read(), '<div id="songlyrics" >', '</div>')
        lyrics = filter_html_tags(lyrics).strip()
        special_chars = {'&#039;': "'"}
        for bad, good in special_chars.items():
            lyrics = lyrics.replace(bad, good)
        if lyrics != '':
            lyrics += '\n\nSource: http://www.lyricstime.com'
        return lyrics

    def get_artist_bio(self, summary=False):
        service = self.services[self.current_station.service_id]
        return service.get_artist_bio(self.current_artist, summary)

    def get_artist_events(self):
        service = self.services[self.current_station.service_id]
        return service.get_artist_events(self.current_artist)

    def get_track_wiki(self, summary=False):
        service = self.services[self.current_station.service_id]
        return service.get_track_wiki(self.current_artist, self.current_track, summary)

    def bookmark_station(self, station):
        bookmarks = self.load('bookmarks')
        if station not in bookmarks:
            bookmarks.append(station)
            self.save('bookmarks', bookmarks)
        self.notify('Station Bookmarked', station.name, 'bookmark-new')

    def debookmark_station(self, station):
        bookmarks = self.load('bookmarks')
        try:
            bookmarks.remove(station)
        except:
            pass
        self.save('bookmarks', bookmarks)
        self.notify('Station Removed from Bookmarks', station.name, 'list-remove')


class ChipmunkApp():

    def __init__(self, glade_file):
        self.still_initting = True
        self.auto_play = False
        self.glade = glade_file
        self.gui = self.get_gui()
        self.chipmunk = Chipmunk()
        self.load_settings()
        self.init_daemons()

    def get_gui(self):
        gui = gtk.Builder()
        gui.add_from_file(self.glade)
        gui.connect_signals(self)
        gui.get_object('expander_context').connect('notify::expanded', self.on_expander_context_notify_expanded)
        gui.get_object('label_expander_context').set_has_tooltip(True)
        gui.get_object('button_cover').set_has_tooltip(True)
        return gui

    def init_daemons(self):
        pynotify.init('Chipmunk Radio')
        dbus.mainloop.glib.DBusGMainLoop(set_as_default=True)
        self.session_bus = dbus.SessionBus()
        self.session_bus.add_signal_receiver(self.on_track_start, signal_name='shellfm_track_started')
        self.bind_keys()
        gobject.timeout_add(1000, self.update_progress)

    def menuitem_station(self, station):
        menu_item = MenuItemStation(station)
        menu_item.connect('activate', self.on_menuitem_station_clicked, station)
        menu_item.show()
        return menu_item

    def toolbutton_station(self, station):
        toolbutton = ToolButtonStation(station)
        toolbutton.connect('clicked', self.on_toolbutton_station_clicked, station)
        toolbutton.set_is_important(True)
        toolbutton.show()
        return toolbutton

    def build_menu(self, menuitem_name, category, recents, context_stations, new_handlers):

        def add_sep(menu, do_it=True):
            if do_it:
                menu_item = gtk.SeparatorMenuItem()
                menu.append(menu_item)
                menu_item.show()

        bookmarks = sorted(self.chipmunk.load('bookmarks'), key=str)
        parent = self.gui.get_object(menuitem_name)
        new_menu = gtk.Menu()

        need_separator = False
        for station in context_stations:
            if station not in bookmarks + recents + [self.chipmunk.current_station]:
                new_menu.append(self.menuitem_station(station))
                need_separator = True
        add_sep(new_menu, need_separator)

        need_separator = False
        for station in bookmarks:
            cat = station.category
            if cat == 'on-the-fly':
                cat = 'group'
            if cat == category:
                new_menu.append(self.menuitem_station(station))
                need_separator = True
        add_sep(new_menu, need_separator)

        need_separator = False
        for station in recents:
            if station not in bookmarks + [self.chipmunk.current_station]:
                new_menu.append(self.menuitem_station(station))
                need_separator = True
        add_sep(new_menu, need_separator)

        #need_separator = False
        #for station in context_stations:
            #if station not in bookmarks + recents + [self.chipmunk.current_station]:
                #new_menu.append(self.menuitem_station(station))
                #need_separator = True
        #add_sep(new_menu, need_separator)

        for text, handler in new_handlers:
            menu_item = gtk.ImageMenuItem(text)
            new_image = gtk.image_new_from_icon_name('list-add', gtk.ICON_SIZE_MENU)
            menu_item.set_image(new_image)
            menu_item.connect('activate', handler)
            new_menu.append(menu_item)
            menu_item.show()

        sub = parent.get_submenu()
        parent.deselect()
        #if sub != None:
            #sub.hide()
        parent.remove_submenu()
        parent.set_submenu(new_menu)

    def build_artists_menu(self):
        menuitem_name = 'menuitem_artists'
        category = 'artist'
        recents = self.chipmunk.load('recent_artists')
        context_stations = []
        artist = self.chipmunk.current_artist.strip()
        if artist != '':
            context_stations.append(self.chipmunk.get_station('artist', artist))
        new_handlers = [('_New Artist', self.on_menuitem_artists_new_activate)]

        self.build_menu(menuitem_name, category, recents, context_stations, new_handlers)

    def build_tags_menu(self, top_tags):
        menuitem_name = 'menuitem_tags'
        category = 'tag'
        recents = self.chipmunk.load('recent_tags')
        context_stations = []
        for tag in top_tags:
            context_stations.append(self.chipmunk.get_station('tag', tag))
        new_handlers = [('_New Tag', self.on_menuitem_tags_new_activate)]

        self.build_menu(menuitem_name, category, recents, context_stations, new_handlers)

    def build_groups_menu(self, group_stations=[]):
        menuitem_name = 'menuitem_groups'
        category = 'group'
        recents = self.chipmunk.load('recent_groups')
        context_stations = group_stations
        new_handlers = [('_New Group', self.on_menuitem_groups_new_activate)]
        new_handlers.append(('New On-the-_Fly Group', self.on_menuitem_groups_rotation_new_activate))

        self.build_menu(menuitem_name, category, recents, context_stations, new_handlers)

    def update_top_tag_toolbutton(self):
        gthreads.AsyncTask(self.chipmunk.get_top_tag,
        self.build_top_tag_toolbutton).start()

    def build_top_tag_toolbutton(self, tag=None):
        widget = self.gui.get_object('toolbutton_top_tag')
        widget.hide()
        if tag != None:
            station = self.chipmunk.get_station('tag', tag)
            if self.chipmunk.current_station != station:
                widget.set_label(station.name)
                widget.show()

    def update_all_menus(self):
        self.update_artists_menu()
        self.update_tags_menu()
        self.update_groups_menu()
        self.update_friends_menu()
        self.update_neighbors_menu()

    def update_menu(self, category):
        menu_map = {'friend': self.update_friends_menu,
        'tag': self.update_tags_menu, 'artist': self.update_artists_menu,
        'group': self.update_groups_menu,
        'on-the-fly': self.update_groups_menu,
        'neighbor': self.update_neighbors_menu}
        menu_map[category]()

    def load_offline_settings(self):
        self.gui.get_object('menuitem_small_icons').set_active(self.chipmunk.load('small_icons'))
        self.gui.get_object('togglebutton_menubar').set_active(self.chipmunk.load('hide_menubar'))
        self.gui.get_object('expander_context').set_expanded(self.chipmunk.load('show_context'))
        self.cover_size = self.chipmunk.load('cover_size')
        self.gui.get_object('image_cover').set_pixel_size(self.cover_size)
        self.label_title_scale = self.chipmunk.load('label_title_scale')
        self.label_artist_scale = self.chipmunk.load('label_artist_scale')
        self.label_album_scale = self.chipmunk.load('label_album_scale')
        self.on_vbox_track_display_scroll_event(None, None)
        self.gui.get_object('window_main').resize(*self.chipmunk.load('window_size'))
        self.gui.get_object('togglebutton_auto_fit').set_active(self.chipmunk.load('auto_fit'))
        self.update_recent_toolbar()
        self.update_artists_menu()
        self.update_groups_menu()
        self.update_tags_menu()
        subscribes = self.chipmunk.load('is_subscriber')
        self.gui.get_object('menuitem_my_loved').set_sensitive(subscribes)
        self.gui.get_object('menuitem_guest_loved').set_sensitive(subscribes)
        #username = self.chipmunk.load('lastfm_username')
        #password = self.chipmunk.load('lastfm_password')

    def load_online_settings(self):
        self.set_subscriber_settings()
        self.update_friends_menu()
        self.update_neighbors_menu()
        self.still_initting = False
        if self.auto_play:
            self.chipmunk.play_pause_track()
        self.auto_play = False

    def connect(self):
        bar = self.gui.get_object('progressbar')
        bar.set_text('Connecting...')
        username = self.chipmunk.load('lastfm_username')
        password = self.chipmunk.load('lastfm_password')
        self.chipmunk.login('lastfm', username, password)
        bar.set_text('Ready to play')

    def load_settings(self):
        self.load_offline_settings()
        gthreads.AsyncTask(self.connect,
        self.load_online_settings).start()

    def on_gnome_mm_key(self, *data):
        bindings = {'Play': self.on_button_play_pause_clicked,
        'Stop': self.toggle_stop_delay,
        'Next': self.on_button_skip_clicked,
        'Previous': self.chipmunk.notify}
        for key in data:
            try:
                bindings[key]()
            except:
                pass

    def bind_keys(self):
        try:
            keybinder.bind('XF86AudioPlay', self.on_button_play_pause_clicked, None)
        except:
            pass

        try:
            keybinder.bind('XF86AudioStop', self.toggle_stop_delay, None)
        except:
            pass

        try:
            keybinder.bind('XF86AudioNext', self.on_button_skip_clicked, None)
        except:
            pass

        try:
            keybinder.bind('XF86AudioPrev', self.chipmunk.notify, None)
        except:
            pass

        self.session_bus.add_signal_receiver(self.on_gnome_mm_key, signal_name='MediaPlayerKeyPressed')

    def toggle_stop_delay(self, data=None):
        widget = self.gui.get_object('togglebutton_stop_delay')
        if widget.props.sensitive:
            widget.set_active(not widget.get_active())

    def bookmark_station(self):
        station = self.chipmunk.current_station
        self.chipmunk.bookmark_station(station)
        self.update_menu(station.category)
        self.gui.get_object('toolbutton_bookmark').hide()
        self.gui.get_object('toolbutton_debookmark').show()

    def debookmark_station(self):
        station = self.chipmunk.current_station
        self.chipmunk.debookmark_station(station)
        self.update_menu(station.category)
        self.gui.get_object('toolbutton_debookmark').hide()
        self.gui.get_object('toolbutton_bookmark').show()

    def add_recent(self, station, category=None, max=10):
        cat_map = {
        'artist': ('recent_artists', self.update_artists_menu),
        'tag': ('recent_tags', self.update_tags_menu),
        'group': ('recent_groups', self.update_groups_menu),
        'on-the-fly': ('recent_groups', self.update_groups_menu),
        None: ('recent_stations', None)}
        if category in cat_map:
            recent_name, recent_update = cat_map[category]
            recent_list = self.chipmunk.load(recent_name)
            try:
                recent_list.remove(station)
            except:
                pass
            recent_list.append(station)
            if len(recent_list) > max:
                recent_list = recent_list[1:]
            self.chipmunk.save(recent_name, recent_list)
            if recent_update != None:
                recent_update()
            if not category:
                self.add_recent(station, station.category)

    def resize(self, image_widget, new_size, image_file=None):
        image_widget.set_pixel_size(new_size)
        if image_file:
            pixbuf = gtk.gdk.pixbuf_new_from_file_at_size(image_file, new_size, new_size)
            image_widget.set_from_pixbuf(pixbuf)
        self.cover_size = new_size


    def on_vbox_track_display_scroll_event(self, widget, event):
        step = 0.1
        min_scale, max_scale = .4, 5.4
        if event == None:
            step = 0
        elif event.direction in [gtk.gdk.SCROLL_DOWN, gtk.gdk.SCROLL_LEFT]:
            step = 0 - step

        old_scales = [self.label_title_scale, self.label_artist_scale, self.label_album_scale]
        #if old_scales[0] == min_scale or old_scales[2] == max_scale:
            #print('reached limit!')
            #step = 0
        new_scales = [scale + step for scale in old_scales]
        widgets = [self.gui.get_object(name) for name in ['label_title', 'label_artist', 'label_album']]
        for label, new_scale in zip(widgets, new_scales):
            if new_scale >= min_scale and new_scale <= max_scale:
                new_attr = pango.AttrScale(new_scale, 0, -1)
                attr_list = label.get_attributes()
                if attr_list != None:
                    attr_list.change(new_attr)
                    label.set_attributes(attr_list)
        self.label_title_scale = new_scales[0]
        self.label_artist_scale = new_scales[1]
        self.label_album_scale = new_scales[2]

    def on_button_cover_scroll_event(self, widget, event):
        image = self.gui.get_object('image_cover')
        old_size = image.get_pixel_size()
        new_size = old_size
        min_size, max_size = 64, 352
        size_step = 32
        if event.direction == gtk.gdk.SCROLL_UP:
            if old_size < max_size:
                new_size += size_step
        elif event.direction == gtk.gdk.SCROLL_DOWN:
            if old_size > min_size:
                new_size -= size_step
        if image.get_icon_name()[0]:
            image_file = None
        else:
            image_file = os.path.join(self.chipmunk.config_dir, 'cover.jpg')
        self.resize(image, new_size, image_file)

    def on_button_new_station_cancel_clicked(self, widget):
        self.gui.get_object('dialog_new_station').hide()

    def on_button_new_station_play_clicked(self, widget):
        if self.next_station_type == 'rotation':
            type = 'library'
            widget = self.gui.get_object('entry_new_station')
            subject = [user.strip() for user in widget.get_text().split(',') if user.strip() != '']
        else:
            type = self.next_station_type
            subject = self.gui.get_object('entry_new_station').get_text()
        self.next_station = self.chipmunk.get_station(type, subject)
        self.gui.get_object('dialog_new_station').hide()
        self.play_station(self.next_station)

    def on_menuitem_artists_new_activate(self, widget):
        self.new_station('lastfm', 'artist')

    def new_station(self, service_id, type):
        #types: library, recommended, neighborhood, loved, artist, tag, group
        self.next_station_service_id = service_id
        self.next_station_type = type
        label_map = {}
        label_tip_map = {}
        for t in ['library', 'recommended', 'neighborhood', 'loved']:
            label_map[t] = 'Username:  '
        label_map['artist'] = 'Artist:  '
        label_map['rotation'] = 'Usernames:  '
        label_tip_map['rotation'] = 'e.g. "EvilGnomeAndy, EvilGnomeCandy, Rockijosh" (without quotes)'
        label_map['group'] = 'Group:  '
        label_map['tag'] = 'Tag:  '
        self.gui.get_object('label_new_station').set_text(label_map[type])
        widget = self.gui.get_object('label_new_station_tip')
        try:
            widget.set_text(label_tip_map[type])
            widget.show()
        except:
            widget.hide()
        self.gui.get_object('entry_new_station').grab_focus()
        self.gui.get_object('dialog_new_station').show()

    def on_button_fit_clicked(self, widget):
        self.fit_window_to_text()

    def on_togglebutton_auto_fit_toggled(self, widget):
        auto_fit = widget.get_active()
        labels = ['label_title', 'label_album', 'label_artist']
        self.ellipsize(labels, not auto_fit)
        old_text = widget.get_tooltip_text()
        self.gui.get_object('window_main').set_resizable(not auto_fit)
        self.chipmunk.save('auto_fit', auto_fit)
        tooltip = {True: "Don't Keep Window Fit to Content", False: 'Keep Window Fit to Content'}
        widget.set_tooltip_text(tooltip[auto_fit])

#THESE ARE OBVIOUSLY REPETITIVE - improve code!
    def on_menuitem_my_recommendations_activate(self, widget):
        username = self.chipmunk.get_username()
        station = self.chipmunk.get_station('recommended', username)
        self.play_station(station)

    def on_menuitem_my_neighborhood_activate(self, widget):
        username = self.chipmunk.get_username()
        station = self.chipmunk.get_station('neighborhood', username)
        self.play_station(station)

    def on_menuitem_my_loved_activate(self, widget):
        username = self.chipmunk.get_username()
        station = self.chipmunk.get_station('loved', username)
        self.play_station(station)

    def on_menuitem_guest_library_activate(self, widget):
        self.new_station('lastfm', 'library')

    def on_menuitem_guest_recommendations_activate(self, widget):
        self.new_station('lastfm', 'recommended')

    def on_menuitem_guest_neighborhood_activate(self, widget):
        self.new_station('lastfm', 'neighborhood')

    def on_menuitem_guest_loved_activate(self, widget):
        self.new_station('lastfm', 'loved')

    def on_menuitem_tags_new_activate(self, widget):
        self.new_station('lastfm', 'tag')

    def on_menuitem_groups_new_activate(self, widget):
        self.new_station('lastfm', 'group')

    def on_menuitem_groups_rotation_new_activate(self, widget):
        self.new_station('lastfm', 'rotation')

    def shrink_window(self, dimension=None):
        widget = self.gui.get_object('window_main')
        width, height = widget.get_size()
        if not dimension:
            width, height = 1, 1
        elif dimension == 'v':
            width, height = width, 1
        elif dimension == 'h':
            width, height = 1, height
        widget.resize(width, height)

    def ellipsize(self, label_names, truncate=True):
        mode_map = {True: pango.ELLIPSIZE_END, False: pango.ELLIPSIZE_NONE}
        for label in [self.gui.get_object(name) for name in label_names]:
            label.set_ellipsize(mode_map[truncate])
        return False

    def fit_window_to_text(self):
        labels = ['label_title', 'label_album', 'label_artist']
        self.ellipsize(labels, False)
        self.shrink_window()
        glib.idle_add(self.ellipsize, labels)

    def clear_progress(self):
        widget = self.gui.get_object('progressbar')
        widget.set_fraction(0)
        widget.set_text('')
        widget.set_tooltip_text('Time Remaining')

    def gtk_widget_hide_on_delete(self, widget, event):
        widget.hide()
        return True

    def update_progress(self):
        widget = self.gui.get_object('progressbar')
        if self.chipmunk.status == 'playing':
            #total_seconds = self.chipmunk.get_total_seconds()
            remaining_seconds = self.chipmunk.get_remaining_seconds()
            if remaining_seconds <= self.chipmunk.time and remaining_seconds > 0:
                widget.set_fraction(remaining_seconds/self.chipmunk.time)
                widget.set_text(str(remaining_seconds // 60) + ':' + str(remaining_seconds % 60).zfill(2))
            else:#does clear_progress cause trouble when it's here and somewhere else?
                widget.set_fraction(0)
                widget.set_text('')
        return True

    def on_window_main_destroy(self, widget):
        self.chipmunk.save('window_size', widget.get_size())
        self.chipmunk.save('label_title_scale', self.label_title_scale)
        self.chipmunk.save('label_artist_scale', self.label_artist_scale)
        self.chipmunk.save('label_album_scale', self.label_album_scale)
        self.chipmunk.save('cover_size', self.cover_size)
        self.chipmunk.disconnect()
        gtk.main_quit()

    def on_button_play_pause_clicked(self, widget):
        self.button_play_pause_toggle()
        if self.still_initting:
            self.auto_play = not self.auto_play
        elif not self.chipmunk.play_pause_track():#try 1
            if not self.chipmunk.play_pause_track():#try 2
                if not self.chipmunk.play_pause_track():#try 3
                    self.button_play_pause_toggle()
                    self.login(self.chipmunk.current_station.service_id)

    def query_delay_station(self, station):
        self.gui.get_object('messagedialog_delay').show()

    def play_station(self, station):
        self.next_station = station
        if self.chipmunk.status == 'playing':
            self.query_delay_station(station)
        elif not self.chipmunk.play_station(station):
            self.login(station.service_id)
        elif self.chipmunk.status == 'paused':
            self.chipmunk.play_pause_track()

    def on_menuitem_station_clicked(self, widget, station):
        self.play_station(station)

    def button_play_pause_toggle(self):
        toggle = {'media-playback-start': 'media-playback-pause',
        'media-playback-pause': 'media-playback-start',
        'Play': 'Pause', 'Pause': 'Play'}
        image = self.gui.get_object('image_play_pause')
        name, size = image.get_icon_name()
        image.set_from_icon_name(toggle[name], size)
        button = self.gui.get_object('button_play_pause')
        button.set_tooltip_text(toggle[button.get_tooltip_text()])

    def on_messagedialog_delay_response(self, widget, response):
        NO, YES, X = -9, -8, -4
        if response == YES:
            self.clear_queue()
            self.chipmunk.queue_station(self.next_station)
            self.gui.get_object('image_queued').show()
            self.gui.get_object('label_queued').set_text(self.next_station.name)
            self.gui.get_object('button_clear_queue').show()
            self.chipmunk.notify('Switch After Track',
            'Chipmunk will switch stations when this track ends', 'appointment-new')
        elif response == NO:
            self.clear_queue()
            if self.chipmunk.status == 'playing':
                self.chipmunk.play_pause_track()
            self.play_station(self.next_station)
        widget.hide()
        #self.gui.get_object('messagedialog_delay').hide()

    def on_togglebutton_scrobble_toggled(self, widget):
        if type(self.chipmunk.services['lastfm'].network) != type(None):
            old_text = widget.get_tooltip_text()
            if widget.get_active():
                widget.set_tooltip_text(old_text.replace('On', 'Off'))
                #turn scrobbling on
                if not self.chipmunk.toggle_scrobble():
                    self.chipmunk.toggle_scrobble()
            else:
                widget.set_tooltip_text(old_text.replace('Off', 'On'))
                #turn scrobbling off
                if self.chipmunk.toggle_scrobble():
                    self.chipmunk.toggle_scrobble()
        #else:
            #widget.set_active(not widget.get_active())

    def on_togglebutton_discovery_toggled(self, widget, stop_recursion=False):
        if type(self.chipmunk.services['lastfm'].network) != type(None):
            old_text = widget.get_tooltip_text()
            if widget.get_active():
                widget.set_tooltip_text(old_text.replace('On', 'Off'))
                #turn discovery on
                if not self.chipmunk.toggle_discovery_mode():
                    self.chipmunk.toggle_discovery_mode()
            else:
                widget.set_tooltip_text(old_text.replace('Off', 'On'))
                #turn discovery off
                if self.chipmunk.toggle_discovery_mode():
                    self.chipmunk.toggle_discovery_mode()
        #else:
            #widget.set_active(not widget.get_active())

    def on_togglebutton_mini_toggled(self, widget):
        #bulky = ['hbox_track_display', 'vbox_menu_toolbar', 'hbox_status_bar',
        #'hseparator']
        bulky = ['hbox_track_display', 'menubar', 'hbox_status_bar', 'hseparator']
        if widget.get_active():
            for w in bulky:
                self.gui.get_object(w).hide()
        else:
            for w in bulky:
                self.gui.get_object(w).show()
            if self.chipmunk.load('hide_menubar'):
                self.hide_menubar()
        self.shrink_window('v')

    def on_togglebutton_menubar_toggled(self, widget):
        self.hide_menubar(widget.get_active())

    def hide_menubar(self, hide=True):
        widget = self.gui.get_object('togglebutton_menubar')
        #image = self.gui.get_object('image_menubar')
        #name, size = image.get_icon_name()
        old_text = widget.get_tooltip_text()
        menubar = self.gui.get_object('menubar')
        if hide:
            widget.set_tooltip_text(old_text.replace('Hide', 'Show'))
            menubar.hide()
            self.chipmunk.save('hide_menubar', True)
            #name = 'go-top'
            self.shrink_window('v')
        else:
            widget.set_tooltip_text(old_text.replace('Show', 'Hide'))
            menubar.show()
            #name = 'go-bottom'
            self.chipmunk.save('hide_menubar', False)
        #image.set_from_icon_name(name, size)

    def on_toolbutton_new_station_inline_clicked(self, widget):
        self.next_station_service_id = 'lastfm'
        for w in ['toolbar_recent_stations', 'toolbar_context']:
            self.gui.get_object(w).hide()
        widget.hide()
        entry = self.gui.get_object('entry_new_station_inline')
        entry.set_text(self.gui.get_object('entry_new_station').get_text())
        for w in ['entry_new_station_inline',
        'toolbutton_new_station_inline_cancel', 'toolbutton_new_group',
        'toolbutton_new_artist', 'toolbutton_new_otf_group',
        'toolbutton_new_guest_library', 'toolbutton_new_tag']:
            self.gui.get_object(w).show()
        entry.grab_focus()

    def on_toolbutton_new_station_inline_cancel_clicked(self, widget):
        for w in ['toolbar_context', 'toolbar_recent_stations',
        'toolbutton_new_station_inline']:
            self.gui.get_object(w).show()
        for w in ['entry_new_station_inline',
        'toolbutton_new_station_inline_cancel', 'toolbutton_new_group',
        'toolbutton_new_artist', 'toolbutton_new_otf_group',
        'toolbutton_new_guest_library', 'toolbutton_new_tag']:
            self.gui.get_object(w).hide()

    def on_toolbutton_new_tag_clicked(self, widget):
        self.on_toolbutton_new_clicked('tag')

    def on_toolbutton_new_clicked(self, next_type):
        self.on_toolbutton_new_station_inline_cancel_clicked(None)
        self.next_station_type = next_type
        subject = self.gui.get_object('entry_new_station_inline').get_text()
        self.gui.get_object('entry_new_station').set_text(subject)
        self.on_button_new_station_play_clicked(None)

    def on_toolbutton_new_artist_clicked(self, widget):
        self.on_toolbutton_new_clicked('artist')

    def on_toolbutton_new_otf_group_clicked(self, widget):
        self.on_toolbutton_new_clicked('rotation')

    def on_toolbutton_new_guest_library_clicked(self, widget):
        self.on_toolbutton_new_clicked('library')

    def on_toolbutton_new_group_clicked(self, widget):
        self.on_toolbutton_new_clicked('group')

    def clear_queue(self):
        self.chipmunk.clear_queue()
        self.gui.get_object('image_queued').hide()
        self.gui.get_object('label_queued').set_text('')
        self.gui.get_object('button_clear_queue').hide()
        self.gui.get_object('togglebutton_stop_delay').set_active(False)

    def on_togglebutton_stop_delay_toggled(self, widget):
        old_text = widget.get_tooltip_text()
        if widget.get_active():
            self.chipmunk.queue_stop()
            widget.set_tooltip_text("Don't " + old_text)
            self.gui.get_object('image_queued').show()
            self.gui.get_object('label_queued').set_text('Stop')
            self.gui.get_object('button_clear_queue').show()
            self.chipmunk.notify('Stop After Track',
            'Chipmunk will stop when this track ends', 'appointment-new')
        else:
            self.clear_queue()
            widget.set_tooltip_text(old_text.replace("Don't ", ''))

    def set_track_labels(self, title='', artist='', album=''):
        l_title = self.gui.get_object('label_title')
        l_artist = self.gui.get_object('label_artist')
        l_album = self.gui.get_object('label_album')
        l_title.set_text(title)
        l_artist.set_text(artist)
        l_album.set_text(album)
        if title:
            l_title.set_tooltip_text('Title: ' + title)
        else:
            l_title.set_tooltip_text('')
        if artist:
            l_artist.set_tooltip_text('Artist: ' + artist)
        else:
            l_artist.set_tooltip_text('')
        if album:
            l_album.set_tooltip_text('Album: ' + album)
        else:
            l_album.set_tooltip_text('')

    def set_cover_image_from_file(self, image_file=None):
        widget = self.gui.get_object('image_cover')
        if image_file == None:
            try:
                station_image = self.chipmunk.current_station.image
            except:
                station_image = None
            if not station_image:
                widget.set_from_icon_name('media-optical', self.cover_size)
                self.tooltip_cover = None
            else:
                pass
        else:
            self.tooltip_cover = gtk.gdk.pixbuf_new_from_file_at_size(image_file, 500, 500)
            #self.tooltip_cover = gtk.gdk.pixbuf_new_from_file(image_file)
        self.resize(widget, self.cover_size, image_file)

    def on_button_cover_query_tooltip(self, widget, x, y, kb_mode, tooltip):
        if not widget.props.sensitive:
            return False
        #the text is annoying next to the pic, I may have to make a custom window
        #to put it below...
        #tooltip.set_text("View Album on Last.fm\nScroll to Resize")
        try:
            tooltip.set_icon(self.tooltip_cover)
        except:
            tooltip.set_icon(None)
        return True

    def on_label_expander_context_query_tooltip(self, widget, x, y, kb_mode, tooltip):
        return self.on_button_cover_query_tooltip(self.gui.get_object('button_cover'), x, y, kb_mode, tooltip)

    def set_cover_image(self):
        gthreads.AsyncTask(self.chipmunk.get_cover_image,
        self.set_cover_image_from_file).start()

    def process_queue(self):
        if not self.chipmunk.process_queue():
            return False
        self.clear_queue()
        if self.chipmunk.status == 'stopped':
            self.on_track_stop()
        return True

    def on_toolbutton_current_artist_clicked(self, widget):
        artist = self.chipmunk.current_artist
        station = self.chipmunk.get_station('artist', artist)
        self.play_station(station)

    def on_toolbutton_top_tag_clicked(self, widget):
        gthreads.AsyncTask(self.chipmunk.get_top_tag,
        self.play_tag).start()
        tag = self.chipmunk.get_top_tag()
        station = self.chipmunk.get_station('tag', tag)
        self.play_station(station)

    def play_tag(self, tag):
        station = self.chipmunk.get_station('tag', tag)
        self.play_station(station)

    def load_all_tabs(self):
        self.load_lyrics_tab()
        self.load_artist_tab()
        self.load_track_tab()
        self.load_events_tab()

    def on_track_start(self, sender=None):
        #TODO: Clean Up!
        #print('track start!')
        self.chipmunk.status = 'playing'
        if not self.process_queue():
            self.chipmunk.time = self.chipmunk.get_total_seconds()
            station = self.chipmunk.current_station
            username = self.chipmunk.get_username()
            title, artist, album = self.chipmunk.get_track(), self.chipmunk.get_artist(), self.chipmunk.get_album()
            #self.chipmunk.play_file('"{p}/{a} - {t}.mp3"'.format(p=self.chipmunk.config_dir, a=artist, t=title))
            self.chipmunk.current_track, self.chipmunk.current_artist = title, artist
            self.chipmunk.cover_image_url = self.chipmunk.get_cover_image_url()
            #force the play/pause button into pause mode if necessary
            widget = self.gui.get_object('image_play_pause')
            if not widget.get_icon_name()[0] == 'media-playback-pause':
                self.button_play_pause_toggle()
            #start updating widgets
            self.set_track_labels(title, artist, album)
            label = self.gui.get_object('label_current_station')
            label.set_text(station.category.title()+': '+station.name)
            for widget in ['button_skip', 'button_share', 'button_tag',
            'button_ban', 'button_love', 'button_download', 'button_cover',
            'togglebutton_stop_delay', 'volumebutton', 'button_share',
            'togglebutton_scrobble', 'togglebutton_discovery']:
                self.gui.get_object(widget).set_sensitive(True)
            for widget in ['toolbutton_debookmark', 'toolbutton_bookmark']:
                self.gui.get_object(widget).hide()
            self.update_recent_toolbar()
            widget = self.gui.get_object('toolbutton_my_library')
            if station != self.chipmunk.get_station('library', username):
                widget.show()
                self.add_recent(station)
                if station.category in ['artist', 'tag', 'group']:
                    if station in self.chipmunk.load('bookmarks'):
                        self.gui.get_object('toolbutton_bookmark').hide()
                        self.gui.get_object('toolbutton_debookmark').show()
                    else:
                        self.gui.get_object('toolbutton_debookmark').hide()
                        self.gui.get_object('toolbutton_bookmark').show()
            else:
                widget.hide()
            #add top tags for current track to the tag menu and toolbar
            self.update_tags_menu()
            self.update_top_tag_toolbutton()
            #add current artist
            self.update_artists_menu()
            widget = self.gui.get_object('toolbutton_current_artist')
            if station != self.chipmunk.get_station('artist', artist):
                widget.set_label(artist)
                widget.show()
            else:
                widget.hide()
            #time = self.chipmunk.get_total_seconds()
            time = str(self.chipmunk.time // 60) + ':' + str(self.chipmunk.time % 60).zfill(2)
            widget = self.gui.get_object('progressbar')
            widget.set_tooltip_text('Time Remaining, of ' + time)
            if self.chipmunk.current_station.rotate():
                self.chipmunk.queue_station(self.chipmunk.current_station)
            self.chipmunk.notify()
            widget = self.gui.get_object('window_main')
            widget.set_title(' '.join([title, 'by', artist, '- Chipmunk']))
            self.set_cover_image()
            self.load_all_tabs()
            #self.chipmunk.toggle_discovery_mode()


    def on_toolbutton_station_clicked(self, widget, station):
        self.play_station(station)

    def update_recent_toolbar(self):
        toolbar = self.gui.get_object('toolbar_recent_stations')
        for toolbutton in toolbar:
            toolbar.remove(toolbutton)
        recent_stations = self.chipmunk.load('recent_stations')
        for station in recent_stations:
            blacklist = [self.chipmunk.current_station,
            self.chipmunk.get_station('artist', self.chipmunk.current_artist)]
            if station not in blacklist:
                toolbutton = self.toolbutton_station(station)
                toolbar.insert(toolbutton, 0)

    def on_track_stop(self):
        widget = self.gui.get_object('image_play_pause')
        if widget.get_icon_name()[0] == 'media-playback-pause':
            self.button_play_pause_toggle()
        self.set_track_labels()
        self.gui.get_object('window_main').set_title('Chipmunk')
        self.clear_progress()
        self.clear_queue()
        for widget in ['button_download',
        'button_share',
        'button_tag',
        'button_ban',
        'button_love',
        'button_share',
        'button_skip',
        'togglebutton_stop_delay',
        'button_cover',
        'volumebutton']:
            self.gui.get_object(widget).set_sensitive(False)
        self.gui.get_object('toolbutton_my_library').show()
        for widget in ['toolbutton_current_artist',
        'toolbutton_top_tag',
        'scrolledwindow_lyrics',
        'linkbutton_lyrics',
        'scrolledwindow_artist',
        'linkbutton_artist',
        'scrolledwindow_track',
        'linkbutton_track',
        'label_events_loading',
        'scrolledwindow_events',
        'linkbutton_events']:
            self.gui.get_object(widget).hide()
        for tab_id in ['lyrics', 'artist', 'track', 'events']:
            self.gui.get_object('label_'+tab_id+'_tab').set_markup(tab_id.title())
        self.set_cover_image_from_file()

    def on_button_cover_clicked(self, widget):
        if self.chipmunk.get_album():
            webbrowser.open_new_tab(self.chipmunk.get_album_url())
        else:
            webbrowser.open_new_tab(self.chipmunk.get_track_url())

    def on_button_skip_clicked(self, widget):
        self.chipmunk.skip_track()

    def on_button_love_clicked(self, widget):
        self.chipmunk.love_track()
        widget.set_sensitive(False)

    def on_button_download_clicked(self, widget):
        #TODO
        #prevent the file from disappearing
        window = self.gui.get_object('filechooserdialog_download')
        #set the name
        window.show()

    def on_expander_context_notify_expanded(self, widget, data):
        if not widget.props.expanded:
            glib.timeout_add(100, self.shrink_window, 'v')
            self.chipmunk.save('show_context', False)
        else:
            self.chipmunk.save('show_context', True)

    def vbox_from_event_dict(self, event_dict):
        linkbutton = gtk.LinkButton(event_dict['url'], event_dict['title'])
        linkbutton.show()
        text = ''
        info_elements = ['venue', 'headliner', 'begins', 'description']
        for element in info_elements:
            info = event_dict[element]
            if info != '':
                if element in ['headliner']:
                    text += element.title()+': '+info+'\n'
                else:
                    text += info+'\n'
        text = text.strip()
        textview = gtk.TextView()
        textview.get_buffer().set_text(text)
        textview.set_editable(False)
        textview.set_cursor_visible(False)
        textview.set_wrap_mode(gtk.WRAP_WORD)
        textview.show()
        #textview.grab_focus()#
        vbox = gtk.VBox()
        vbox.show()
        vbox.pack_start(linkbutton)
        vbox.pack_start(textview)
        #vbox.set_child_packing(textview, True, True, padding, pack_type)
        #textview.grab_focus()#
        return vbox

    def build_lyrics_tab(self, content):
        self.build_tab('lyrics', content)

    def build_artist_tab(self, content):
        self.build_tab('artist', content)

    def build_track_tab(self, content):
        self.build_tab('track', content)

    def build_events_tab(self, content):
        self.build_tab('events', content)

    def build_tab(self, tab_id, content):
        if content == None:
            content = ''
        elif type(content) == type(''):
            content = content.strip()
        label = self.gui.get_object('label_'+tab_id+'_tab')
        label_text = filter_html_tags(label.get_label())
        scrolled_window = self.gui.get_object('scrolledwindow_'+tab_id)
        linkbutton = self.gui.get_object('linkbutton_'+tab_id)
        if tab_id == 'events':
            for child in scrolled_window.get_children():
                scrolled_window.remove(child)
            if len(content) == 0:
                scrolled_window.hide()
                linkbutton.show()
                label.set_markup_with_mnemonic(label_text)
            else:
                label.set_markup_with_mnemonic('<b>'+label_text+'</b>')
                linkbutton.hide()
                events_vbox = gtk.VBox(spacing=5)
                for event_dict in content:
                    event_vbox = self.vbox_from_event_dict(event_dict)
                    events_vbox.pack_start(event_vbox)
                scrolled_window.add_with_viewport(events_vbox)
                scrolled_window.show()
                events_vbox.show()
            self.gui.get_object('label_events_loading').hide()
        else:
            textview = self.gui.get_object('textview_'+tab_id)
            if content == '' and self.chipmunk.current_track.strip() != '':
                linkbutton.show()
                scrolled_window.hide()
                label.set_markup_with_mnemonic(label_text)
            else:
                try:
                    textview.get_buffer().set_text(unicode(content, errors='replace'))
                except:
                    textview.get_buffer().set_text(content)
                scrolled_window.show()
                linkbutton.hide()
                label.set_markup_with_mnemonic('<b>'+label_text+'</b>')

    def load_lyrics_tab(self):
        gthreads.AsyncTask(self.chipmunk.get_lyrics,
        self.build_lyrics_tab).start()

    def load_artist_tab(self):
        gthreads.AsyncTask(self.chipmunk.get_artist_bio,
        self.build_artist_tab).start()

    def load_track_tab(self):
        gthreads.AsyncTask(self.chipmunk.get_track_wiki,
        self.build_track_tab).start()

    def load_events_tab(self):
        self.gui.get_object('linkbutton_events').hide()
        self.gui.get_object('scrolledwindow_events').hide()
        self.gui.get_object('label_events_loading').show()
        gthreads.AsyncTask(self.chipmunk.get_artist_events,
        self.build_events_tab).start()

    def on_linkbutton_lyrics_clicked(self, widget):
        query = 'lyrics "'+self.chipmunk.current_artist+'" "'+self.chipmunk.current_track+'"'
        query = query.replace(' ', '+')
        query = query.replace('&', '%26')
        query += '&tbs=prv:1'
        webbrowser.open_new_tab('http://www.google.com/search?q=' + query)

    def on_linkbutton_events_clicked(self, widget):
        webbrowser.open_new_tab(self.chipmunk.get_artist_url()+'/+events')

    def on_linkbutton_track_clicked(self, widget):
        webbrowser.open_new_tab(self.chipmunk.get_track_url())

    def on_linkbutton_artist_clicked(self, widget):
        webbrowser.open_new_tab(self.chipmunk.get_artist_url())

    def on_button_share_clicked(self, widget):
        #populate combobox with friends
        friends = self.chipmunk.get_friends()
        widget = self.gui.get_object('comboboxentry_share')
        model = gtk.ListStore(str)
        for friend in friends:
            model.append([friend])
        widget.set_model(model)
        widget.set_text_column(0)
        widget = self.gui.get_object('checkbutton_share_shout')
        widget.set_active(self.chipmunk.load('shout_when_share'))
        self.gui.get_object('dialog_share').show()

    def on_volumebutton_value_changed(self, widget, vol=1):
        vol = int(vol*100)
        self.chipmunk.set_volume(vol)
        return True

    def on_button_share_cancel_clicked(self, widget):
        self.gui.get_object('dialog_share').hide()

    def on_button_share_send_clicked(self, widget):
        track = self.chipmunk.current_track
        entry = self.gui.get_object('comboboxentry_share').child
        user_list = [user.strip() for user in entry.get_text().split(',')]
        entry = self.gui.get_object('entry_share_message')
        message = entry.get_text()
        self.chipmunk.share_track(user_list, message)
        if self.gui.get_object('checkbutton_share_shout').get_active():
            self.chipmunk.save('shout_when_share', True)
            self.chipmunk.shout_track(user_list, message)
        else:
            self.chipmunk.save('shout_when_share', False)
        self.chipmunk.notify('Track Shared',
        track+' sent to '+', '.join(user_list), 'emblem-shared')
        self.gui.get_object('dialog_share').hide()

    def on_button_tag_clicked(self, widget):
        #TODO
        pass

    def on_button_ban_clicked(self, widget):
        self.chipmunk.ban_track()

    def on_button_clear_queue_clicked(self, widget):
        self.clear_queue()

    def on_toolbutton_bookmark_clicked(self, widget):
        station = self.chipmunk.current_station
        self.bookmark_station()

    def on_toolbutton_debookmark_clicked(self, widget):
        self.debookmark_station()

    def login_helper(self, service_id, username, password):
        self.chipmunk.login(service_id, username, password)
        self.set_subscriber_settings()
        self.update_friends_menu()
        self.update_neighbors_menu()

    def login(self, service_id='lastfm'):
        stored_username = self.chipmunk.load(service_id + '_username')
        stored_password = self.chipmunk.load(service_id + '_password')
        widget = self.gui.get_object('entry_login_username')
        widget.set_text(stored_username)
        widget.grab_focus()
        self.gui.get_object('entry_login_password').set_text(stored_password)
        self.gui.get_object('dialog_login').show()

    def on_menuitem_manage_lastfm_activate(self, widget):
        self.login('lastfm')

    def on_button_login_cancel_clicked(self, widget):
        self.gui.get_object('dialog_login').hide()

    def on_button_login_ok_clicked(self, widget):
        username = self.gui.get_object('entry_login_username').get_text()
        password = self.gui.get_object('entry_login_password').get_text()

        if username and password:
            self.login_helper('lastfm', username, password)
        self.gui.get_object('dialog_login').hide()
        self.chipmunk.notify('Login and Password Set',
        'Click the play button or choose a station to begin', 'dialog-password')

    def build_friends_menu(self, users):
        menuitem_name = 'menuitem_friends'
        category = 'friend'
        parent = self.gui.get_object(menuitem_name)
        new_menu = gtk.Menu()
        for user in users:
            station = self.chipmunk.get_station('library', user, category)
            new_menu.append(self.menuitem_station(station))
        parent.remove_submenu()
        parent.set_submenu(new_menu)

    def build_neighbors_menu(self, users):
        menuitem_name = 'menuitem_neighbors'
        category = 'neighbor'
        parent = self.gui.get_object(menuitem_name)
        new_menu = gtk.Menu()
        for user in users:
            station = self.chipmunk.get_station('library', user, category)
            new_menu.append(self.menuitem_station(station))
        parent.remove_submenu()
        parent.set_submenu(new_menu)

    def update_artists_menu(self):
        self.build_artists_menu()

    def update_tags_menu(self):
        gthreads.AsyncTask(self.chipmunk.get_top_tags,
        self.build_tags_menu).start()

    def update_groups_menu(self):
        self.build_groups_menu()

    def update_friends_menu(self):
        gthreads.AsyncTask(self.chipmunk.get_friends,
        self.build_friends_menu).start()

    def update_neighbors_menu(self):
        gthreads.AsyncTask(self.chipmunk.get_neighbors,
        self.build_neighbors_menu).start()

    def set_subscriber_settings(self, subscribes=None):
        if subscribes == None:
            gthreads.AsyncTask(self.chipmunk.is_subscriber,
            self.set_subscriber_settings).start()
        else:
            self.gui.get_object('menuitem_my_loved').set_sensitive(subscribes)
            self.gui.get_object('menuitem_guest_loved').set_sensitive(subscribes)
            self.chipmunk.save('is_subscriber', subscribes)

    def on_menuitem_manage_development_activate(self, widget):
        webbrowser.open_new_tab('http://bitbucket.org/EvilGnome/chipmunk/issues/')

    def on_menuitem_small_icons_toggled(self, widget):
        large = ['image_play_pause', 'image_stop_delay', 'image_mini', 'image_skip']
        small = ['image_love', 'image_share', 'image_ban']
        #small = ['image_love', 'image_share', 'image_ban', 'image_download', 'image_tag']
        shrink = widget.get_active()
        new_size = {True: (36, 24), False: (48, 36)}[shrink]
        for img in large:
            self.gui.get_object(img).set_pixel_size(new_size[0])
        for img in small:
            self.gui.get_object(img).set_pixel_size(new_size[1])
        self.chipmunk.save('small_icons', shrink)

    def on_menuitem_somafm_website_activate(self, widget):
        webbrowser.open_new_tab('http://somafm.com/')

    def on_toolbutton_my_library_clicked(self, widget):
        username = self.chipmunk.get_username()
        station = self.chipmunk.get_station('library', username)
        self.play_station(station)


if __name__ == '__main__':
    #Ensure single instance
    pid_file = '/tmp/chipmunk-radio.pid'
    fp = open(pid_file, 'w')
    try:
        fcntl.lockf(fp, fcntl.LOCK_EX | fcntl.LOCK_NB)
    except IOError:
        sys.exit(0)

    app = ChipmunkApp('chipmunk-radio.glade')
    gtk.main()
